package producerconsumercanvas;

import java.util.concurrent.ConcurrentLinkedQueue;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.concurrent.Task;
import javafx.geometry.Insets;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.layout.Background;
import javafx.scene.layout.BackgroundFill;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.CornerRadii;
import javafx.scene.paint.Color;
import javafx.stage.Stage;

/**
 * Java SE 8 and up compatible example for processing significant
 * amounts of data points such that the points can be transformed and
 * graphed onto a JavaFX Canvas in the order that they arrive.
 *
 * @author sphillips
 */
public class ProducerConsumerCanvas extends Application {

    Canvas canvas; //For plotting our data
    ConcurrentLinkedQueue<PointPojo> pointQueue; 

    int setSize = 1000000; //Total number of points to plot
    double radius = 3.0; //The width of each Canvas oval
    //Predefining data ranges simplifies coordinate transformations 
    //Data will range from -1.0 to 1.0
    double totalMinX = -1.0; 
    double totalRangeX = 2.0;
    double totalMinY = -1.0;
    double totalRangeY = 2.0;

    @Override
    public void start(Stage primaryStage) {
        //Setup simple scene and layout for demonstration
        canvas = new Canvas(1000, 1000);
        BorderPane bp = new BorderPane(canvas);
        bp.setBackground(
            new Background(
                new BackgroundFill(
                    Color.BLACK, CornerRadii.EMPTY, Insets.EMPTY)));
        Scene scene = new Scene(bp, Color.BLACK);

        primaryStage.setTitle("Producer Consumer Canvas Example");
        primaryStage.setScene(scene);
        primaryStage.show();

        //Thread safe Queue used for data exchange
        pointQueue = new ConcurrentLinkedQueue<>();

        //Define producing task to create and add data points to queue
        Task<Void> producerTask = new Task<Void>() {
            @Override
            protected Void call() throws Exception {
                int count = 0; //track how many times we have produced data
                //Produce data asynchronously forever
                while (true) {
                    //wait until the data has been processed 
                    if (!pointQueue.isEmpty()) {
                        Thread.sleep(0, 100); //Spin the thread a bit
                    } else {
                        //The queue is ready to be filled again
                        System.out.println("Producer Task run: " + count);
                        for (int i = 0; i < setSize; i++) {
                            PointPojo p = new PointPojo();
                            pointQueue.add(p);
                        }
                        count++;  
                    }
                }
            }
        };
        Thread producerThread = new Thread(producerTask);
        producerThread.setDaemon(true);

        //Start consumer task for rendering the data
        Task<Void> consumerTask = new Task<Void>() {
            @Override
            protected Void call() throws Exception {
                resetCanvas(); //randomize current canvas fill color
                //track sets of points that are plotted
                int plotCount = 0;
                while (true) {
                    //drawing in batches reduces thread contention
                    long startTime = System.nanoTime();
                    //int batchSize = drawNext_SimplePolling();
                    int batchSize = drawNext_ArrayTransforms();
                    printTotalTime(startTime);
                    System.out.println("Drew batch size of " + batchSize);
                    plotCount += batchSize;
                    //little thread spin helps it animate 
                    //less than 15 ms can freeze graphics
                    Thread.sleep(25);
                    //if the set is plotted lets change the colors
                    if (plotCount >= setSize) {
                        resetCanvas();
                        plotCount = 0;
                        System.out.println("Reset Canvas");
                    }
                }
            }
        };

        Thread consumerThread = new Thread(consumerTask);
        consumerThread.setDaemon(true);
        //start polling the Queue
        consumerThread.start();
        //with the consumer thread polling, start feeding Queue
        producerThread.start();
    }

    private int drawNext_SimplePolling() {
        int size = pointQueue.size();
        GraphicsContext g = canvas.getGraphicsContext2D();
        Platform.runLater(() -> {
            while (!pointQueue.isEmpty()) {
                PointPojo point = pointQueue.poll();
                //coordinate transformation from data to canvas pixels
                double x = transformXToScreen(point.x);
                double y = transformYToScreen(point.y);
                //encourage the object to be garbage collected sooner
                point = null;  
                g.fillOval(x, y, radius, radius);
            }
        });
        return size;
    }

private int drawNext_ArrayTransforms() {
    int size = pointQueue.size();
    GraphicsContext g = canvas.getGraphicsContext2D();
    //temporary arrays to hold the transformed canvas coordinates
    //Be sure to use primitive double type to minimize memory
    double[] xArray = new double[size];
    double[] yArray = new double[size];
    //using an iterator takes over an order of magnitude longer 
    //than polling because the Hotspot has not yet "warmed up"
    for (int i = 0; i < size; i++) {
        PointPojo point = pointQueue.poll();
        //coordinate transformations stored in temporary arrays
        xArray[i] = transformXToScreen(point.x);
        yArray[i] = transformXToScreen(point.y);
        //encourage finalization of the object we polled
        point = null; 
    }
    //Using arrays to hold transformed coordinates allows us 
    //to use a runLater() thread while minimizing blocking time
    Platform.runLater(() -> {
        //The key is minimizing time spent in this blocking thread
        for (int i = 0; i < size; i++) {
            g.fillOval(xArray[i], yArray[i], radius, radius);
        }
    });
    return size;
}

    private void resetCanvas() {
        System.out.println("Resetting Consumer Canvas...");
        //Obtain a fresh reference to the Canvas graphics context
        GraphicsContext g = canvas.getGraphicsContext2D();
        //Change color for canvas painting
        g.setFill(Color.hsb(Math.random() * 360.0, 1.0, 1.0, 0.75));
//        Platform.runLater(()->{
//            //Clear the screen
//            g.clearRect(0, 0, canvas.getWidth(), canvas.getHeight());
//        });
    }

    private double transformXToScreen(double x) {
        return (((x - totalMinX) * (canvas.getWidth() - radius)) / totalRangeX) + radius;
    }

    private double transformYToScreen(double y) {
        return (((y - totalMinY) * (canvas.getHeight() - radius)) 
            / totalRangeY) + radius;
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        launch(args);
    }

    /**
     * Helper utility for getting sub millisecond timing metrics
     *
     * @param startTime the System.nanoTime() to compare the current
     * time against
     */
    private static void printTotalTime(long startTime) {
        long estimatedTime = System.nanoTime() - startTime;
        long totalNanos = estimatedTime;
        long s = totalNanos / 1000000000;
        totalNanos -= s * 1000000000;
        long ms = totalNanos / 1000000;
        totalNanos -= ms * 1000000;
        long us = totalNanos / 1000;
        totalNanos -= us * 1000;
        System.out.println("Total elapsed time: Total ns: " 
            + estimatedTime + ", " + s + ":s:" + ms + ":ms:" 
            + us + ":us:" + totalNanos + ":ns");
    }
}